import fs from 'fs';
import path from 'path';
import { PNG } from 'pngjs';
import pixelmatch from 'pixelmatch';
import moveFile from 'move-file';
import sharp from 'sharp';

function isHighSurrogate(codePoint) {
  return codePoint >= 0xd800 && codePoint <= 0xdbff;
}

function isLowSurrogate(codePoint) {
  return codePoint >= 0xdc00 && codePoint <= 0xdfff;
}

// Truncate string by size in bytes
var truncate = function truncate(getLength, string, byteLength) {
  if (typeof string !== "string") {
    throw new Error("Input must be string");
  }

  var charLength = string.length;
  var curByteLength = 0;
  var codePoint;
  var segment;

  for (var i = 0; i < charLength; i += 1) {
    codePoint = string.charCodeAt(i);
    segment = string[i];

    if (isHighSurrogate(codePoint) && isLowSurrogate(string.charCodeAt(i + 1))) {
      i += 1;
      segment += string[i];
    }

    curByteLength += getLength(segment);

    if (curByteLength === byteLength) {
      return string.slice(0, i + 1);
    }
    else if (curByteLength > byteLength) {
      return string.slice(0, i - segment.length + 1);
    }
  }

  return string;
};

var getLength = Buffer.byteLength.bind(Buffer);
var truncateUtf8Bytes = truncate.bind(null, getLength);

/*jshint node:true*/

/**
 * Replaces characters in strings that are illegal/unsafe for filenames.
 * Unsafe characters are either removed or replaced by a substitute set
 * in the optional `options` object.
 *
 * Illegal Characters on Various Operating Systems
 * / ? < > \ : * | "
 * https://kb.acronis.com/content/39790
 *
 * Unicode Control codes
 * C0 0x00-0x1f & C1 (0x80-0x9f)
 * http://en.wikipedia.org/wiki/C0_and_C1_control_codes
 *
 * Reserved filenames on Unix-based systems (".", "..")
 * Reserved filenames in Windows ("CON", "PRN", "AUX", "NUL", "COM1",
 * "COM2", "COM3", "COM4", "COM5", "COM6", "COM7", "COM8", "COM9",
 * "LPT1", "LPT2", "LPT3", "LPT4", "LPT5", "LPT6", "LPT7", "LPT8", and
 * "LPT9") case-insesitively and with or without filename extensions.
 *
 * Capped at 255 characters in length.
 * http://unix.stackexchange.com/questions/32795/what-is-the-maximum-allowed-filename-and-folder-size-with-ecryptfs
 *
 * @param  {String} input   Original filename
 * @param  {Object} options {replacement: String | Function }
 * @return {String}         Sanitized filename
 */



var illegalRe = /[\/\?<>\\:\*\|"]/g;
var controlRe = /[\x00-\x1f\x80-\x9f]/g;
var reservedRe = /^\.+$/;
var windowsReservedRe = /^(con|prn|aux|nul|com[0-9]|lpt[0-9])(\..*)?$/i;
var windowsTrailingRe = /[\. ]+$/;

function sanitize(input, replacement) {
  if (typeof input !== 'string') {
    throw new Error('Input must be string');
  }
  var sanitized = input
    .replace(illegalRe, replacement)
    .replace(controlRe, replacement)
    .replace(reservedRe, replacement)
    .replace(windowsReservedRe, replacement)
    .replace(windowsTrailingRe, replacement);
  return truncateUtf8Bytes(sanitized, 255);
}

var sanitizeFilename = function (input, options) {
  var replacement = (options && options.replacement) || '';
  var output = sanitize(input, replacement);
  if (replacement === '') {
    return output;
  }
  return sanitize(output, '');
};

const PLUGIN_NAME = "cp-visual-regression-diff";
const IMAGE_SNAPSHOT_PREFIX = `__${PLUGIN_NAME}_snapshots__`;
var FILE_SUFFIX;

(function (FILE_SUFFIX) {
  FILE_SUFFIX["diff"] = ".diff";
  FILE_SUFFIX["actual"] = ".actual";
})(FILE_SUFFIX || (FILE_SUFFIX = {}));

const TASK = {
  getScreenshotPath: `${PLUGIN_NAME}-getScreenshotPath`,
  compareImages: `${PLUGIN_NAME}-compareImages`,
  approveImage: `${PLUGIN_NAME}-approveImage`,
  doesFileExist: `${PLUGIN_NAME}-doesFileExist`
  /* c8 ignore next */

};

const inArea = (x, y, height, width) => y > height || x > width;

const fillSizeDifference = (image, width, height) => {
  for (let y = 0; y < image.height; y++) {
    for (let x = 0; x < image.width; x++) {
      if (inArea(x, y, height, width)) {
        const idx = image.width * y + x << 2;
        image.data[idx] = 0;
        image.data[idx + 1] = 0;
        image.data[idx + 2] = 0;
        image.data[idx + 3] = 64;
      }
    }
  }

  return image;
  /* c8 ignore next */
};
const createImageResizer = (width, height) => source => {
  const resized = new PNG({
    width,
    height,
    fill: true
  });
  PNG.bitblt(source, resized, 0, 0, source.width, source.height, 0, 0);
  return resized;
  /* c8 ignore next */
};
const importAndScaleImage = async cfg => {
  const imgBuffer = fs.readFileSync(cfg.path);
  const rawImgNew = PNG.sync.read(imgBuffer);
  if (cfg.scaleFactor === 1) return rawImgNew;
  const newImageWidth = Math.ceil(rawImgNew.width * cfg.scaleFactor);
  const newImageHeight = Math.ceil(rawImgNew.height * cfg.scaleFactor);
  await sharp(imgBuffer).resize(newImageWidth, newImageHeight).toFile(cfg.path);
  return PNG.sync.read(fs.readFileSync(cfg.path));
};
const alignImagesToSameSize = (firstImage, secondImage) => {
  const firstImageWidth = firstImage.width;
  const firstImageHeight = firstImage.height;
  const secondImageWidth = secondImage.width;
  const secondImageHeight = secondImage.height;
  const resizeToSameSize = createImageResizer(Math.max(firstImageWidth, secondImageWidth), Math.max(firstImageHeight, secondImageHeight));
  const resizedFirst = resizeToSameSize(firstImage);
  const resizedSecond = resizeToSameSize(secondImage);
  return [fillSizeDifference(resizedFirst, firstImageWidth, firstImageHeight), fillSizeDifference(resizedSecond, secondImageWidth, secondImageHeight)];
};

const round = n => Math.ceil(n * 1000) / 1000;

const unlinkSyncSafe = path => fs.existsSync(path) && fs.unlinkSync(path);

const moveSyncSafe = (pathFrom, pathTo) => fs.existsSync(pathFrom) && moveFile.sync(pathFrom, pathTo);

const getScreenshotPathTask = ({
  title,
  imagesDir,
  specPath
}) => path.join(IMAGE_SNAPSHOT_PREFIX, path.dirname(specPath), ...imagesDir.split("/"), `${sanitizeFilename(title)}${FILE_SUFFIX.actual}.png`);
const approveImageTask = ({
  img
}) => {
  const oldImg = img.replace(FILE_SUFFIX.actual, "");
  unlinkSyncSafe(oldImg);
  const diffImg = img.replace(FILE_SUFFIX.actual, FILE_SUFFIX.diff);
  unlinkSyncSafe(diffImg);
  moveSyncSafe(img, oldImg);
  return null;
};
const compareImagesTask = async cfg => {
  const messages = [];
  let imgDiff;
  let imgNewBase64, imgOldBase64, imgDiffBase64;
  let error = false;

  if (fs.existsSync(cfg.imgOld) && !cfg.updateImages) {
    const rawImgNew = await importAndScaleImage({
      scaleFactor: cfg.scaleFactor,
      path: cfg.imgNew
    });
    const rawImgOld = PNG.sync.read(fs.readFileSync(cfg.imgOld));
    const isImgSizeDifferent = rawImgNew.height !== rawImgOld.height || rawImgNew.width !== rawImgOld.width;
    const [imgNew, imgOld] = isImgSizeDifferent ? alignImagesToSameSize(rawImgNew, rawImgOld) : [rawImgNew, rawImgOld];
    const {
      width,
      height
    } = imgNew;
    const diff = new PNG({
      width,
      height
    });
    const diffConfig = Object.assign({
      includeAA: true
    }, cfg.diffConfig);
    const diffPixels = pixelmatch(imgNew.data, imgOld.data, diff.data, width, height, diffConfig);
    imgDiff = diffPixels / (width * height);

    if (isImgSizeDifferent) {
      messages.push(`Warning: Images size mismatch - new screenshot is ${rawImgNew.width}px by ${rawImgNew.height}px while old one is ${rawImgOld.width}px by ${rawImgOld.height} (width x height).`);
    }

    if (imgDiff > cfg.maxDiffThreshold) {
      messages.unshift(`Image diff factor (${round(imgDiff)}%) is bigger than maximum threshold option ${cfg.maxDiffThreshold}.`);
      error = true;
    }

    const diffBuffer = PNG.sync.write(diff);
    imgNewBase64 = PNG.sync.write(imgNew).toString("base64");
    imgDiffBase64 = diffBuffer.toString("base64");
    imgOldBase64 = PNG.sync.write(imgOld).toString("base64");

    if (error) {
      fs.writeFileSync(cfg.imgNew.replace(FILE_SUFFIX.actual, FILE_SUFFIX.diff), diffBuffer);
      return {
        error,
        message: messages.join("\n"),
        imgDiff,
        imgNewBase64,
        imgDiffBase64,
        imgOldBase64,
        maxDiffThreshold: cfg.maxDiffThreshold
      };
    } else {
      // don't overwrite file if it's the same (imgDiff < cfg.maxDiffThreshold && !isImgSizeDifferent)
      fs.unlinkSync(cfg.imgNew);
    }
  } else {
    // there is no "old screenshot" or screenshots should be immediately updated
    imgDiff = 0;
    imgNewBase64 = "";
    imgDiffBase64 = "";
    imgOldBase64 = "";
    moveFile.sync(cfg.imgNew, cfg.imgOld);
  }

  if (typeof imgDiff !== "undefined") {
    messages.unshift(`Image diff factor (${round(imgDiff)}%) is within boundaries of maximum threshold option ${cfg.maxDiffThreshold}.`);
    return {
      message: messages.join("\n"),
      imgDiff,
      imgNewBase64,
      imgDiffBase64,
      imgOldBase64,
      maxDiffThreshold: cfg.maxDiffThreshold
    };
  }
  /* c8 ignore next */


  return null;
};
const doesFileExistTask = ({
  path
}) => fs.existsSync(path);
/* c8 ignore start */

const initTaskHook = () => ({
  [TASK.getScreenshotPath]: getScreenshotPathTask,
  [TASK.doesFileExist]: doesFileExistTask,
  [TASK.approveImage]: approveImageTask,
  [TASK.compareImages]: compareImagesTask
});
/* c8 ignore stop */

const getConfigVariableOrThrow = (config, name) => {
  if (config[name]) {
    return config[name];
  }
  /* c8 ignore start */


  throw `[Image snapshot] CypressConfig.${name} cannot be missing or \`false\`!`;
};
/* c8 ignore stop */


const removeScreenshotsDirectory = (screenshotsFolder, onSuccess, onError) => {
  fs.rm(path.join(screenshotsFolder, IMAGE_SNAPSHOT_PREFIX), {
    recursive: true,
    force: true
  }, err => {
    /* c8 ignore start */
    if (err) return onError(err);
    /* c8 ignore stop */

    onSuccess();
  });
};

const initAfterScreenshotHook = config => details => {
  var _details$name;

  // it's not a screenshot generated by FRSOURCE Cypress Plugin Visual Regression Diff

  /* c8 ignore start */
  if (((_details$name = details.name) == null ? void 0 : _details$name.indexOf(IMAGE_SNAPSHOT_PREFIX)) !== 0) return;
  /* c8 ignore stop */

  return new Promise((resolve, reject) => {
    const screenshotsFolder = getConfigVariableOrThrow(config, "screenshotsFolder");
    const newRelativePath = details.name.substring(IMAGE_SNAPSHOT_PREFIX.length + path.sep.length);
    const newAbsolutePath = path.normalize(path.join(config.projectRoot, newRelativePath));
    void moveFile(details.path, newAbsolutePath).then(() => removeScreenshotsDirectory(screenshotsFolder, () => resolve({
      path: newAbsolutePath
    }), reject)).catch(reject);
  });
};

/* c8 ignore start */

const initForceDeviceScaleFactor = on => {
  // based on https://github.com/cypress-io/cypress/issues/2102#issuecomment-521299946
  on("before:browser:launch", (browser, launchOptions) => {
    if (browser.name === "chrome" || browser.name === "chromium") {
      launchOptions.args.push("--force-device-scale-factor=1");
      launchOptions.args.push("--high-dpi-support=1");
    } else if (browser.name === "electron" && browser.isHeaded) {
      // eslint-disable-next-line no-console
      console.log("There isn't currently a way of setting the device scale factor in Cypress when running headed electron so we disable the image regression commands.");
    }
  });
};
/* c8 ignore stop */


const initPlugin = (on, config) => {
  /* c8 ignore start */
  if (config.env["pluginVisualRegressionForceDeviceScaleFactor"] !== false) {
    initForceDeviceScaleFactor(on);
  }
  /* c8 ignore stop */


  on("task", initTaskHook());
  on("after:screenshot", initAfterScreenshotHook(config));
};

export { initPlugin };
//# sourceMappingURL=plugins.mjs.map
